
#ifdef USE_DCS_DEFERRED

//#define FINALCOLOR_DISABLE_DEPTH_TEST		- отключает depth test
//#define FINALCOLOR_DISABLE_ALPHA_BLEND    - отключает альфа блендинг
#include "inc/StencilUsage.hlsl"
#include "inc/CullMode.hlsl"
#include "inc/DepthFunc.hlsl"

#ifndef FOG_ENABLE
	#define FOG_ENABLE
#endif

#ifndef CLOUD_SHADOWS
	#define CLOUD_SHADOWS
#endif

#ifndef LIGHT_TEXTURE
	#define LIGHT_TEXTURE
#endif

#ifndef METASHADER
	#define METASHADER
#endif

#include "inc/DcsForward.hlsl"

#ifdef CUSTOM_DEPTH
	#include "../lockonshaders/common/customDepth.hlsl"	
#endif

// #define SURFACELOD
// #define NEXTLOD_BLENDING
#include "terrain/inc/surfaceLodsBlendingConfig.hlsl"
#include "terrain/inc/surfaceLodsBlending.hlsl"

#include "../deferred/_HMD.hlsl"
#include "../deferred/_helper.hlsl"

void DCS_Forward_ps_preprocessor(
	float4 positionWS : COLOR0,
#ifdef CLOUD_SHADOWS
	float4 cloudShadowsUV : COLOR1,
#endif
	float4 sv_position : SV_POSITION,
#ifdef CUSTOM_DEPTH
	out float depth : SV_DEPTH,
#endif
	out float4 color : SV_Target
#if defined(THIS_IS_TREE) && ENABLE_ATOC
  ,out uint coverageMask : SV_Coverage
#endif  
PS_IN_DECLARATION
	)
{
	float4 NDCposition = mul(positionWS, TerrainContext_VP);

#if defined(SURFACE5) && defined(NEXTLOD_CLIPPING) && defined(USE_blendSurfaceClippingPS)
	float alphaDistance = 0;
	int visible = blendSurfaceClippingPS(positionWS, alphaDistance);
	//color = float4(visible*0.5, 0, 0, 1); return;
	if (visible == 0)
		discard;

#endif

	// дистанция
	float dist = length( TerrainContext_cameraPos-positionWS.xyz);
#ifdef OLD_LOD0_BLENDING
	float distanceBlend = TerrainContext_surfaceLevel0*(1-transitionDistance);
	float alphaDistance = lerp( 0, 1, saturate( (TerrainContext_surfaceLevel0-dist)/distanceBlend ));
	if(alphaDistance==0)
		discard;
#endif

#ifdef OLD_LOD1_BLENDING
	float distanceBlend = transitionDistance*TerrainContext_surfaceLevel0;
	float alphaDistance = step( distanceBlend, dist );
	if(alphaDistance==0)
		discard;
	//color = float4(alphaDistance, alphaDistance, alphaDistance, 1); return;
#endif

#ifdef CUSTOM_DEPTH
	float zNear = TerrainContext_P[3][2] / (TerrainContext_P[2][2]-1.0);
	float zFar = TerrainContext_P[3][2] / (1.0+TerrainContext_P[2][2]);
	depth=CustomDepth(NDCposition, zNear, zFar);
#endif

	ColorOutput output;
	ColorOutput_Init(output);
	// вызов пиксельного шейдера
	PS_NAME(
		output, 
		positionWS 
		PS_IN_CALL);

#if defined(THIS_IS_TREE) && ENABLE_ATOC
  if (COVERAGE_MASK > 1)
  {
    coverageMask = AToC(output.fAlpha);  
    output.fAlpha = 1;
  }
#endif

#if defined(RENDER_VARYHOUSE_COMPONENTS) && !defined(THIS_IS_VARYHOUSE_MATERIAL)
	discard;
#endif
#if defined(RENDER_SURFACE_COMPONENTS) && !defined(THIS_IS_SURFACE_MATERIAL)
	discard;
#endif


	color = DCS_Forward(
		output, 
		positionWS,
		NDCposition,
		sv_position
	);

}

technique10 DCS_Forward_Technique
{
	pass P0
	{
		SetVertexShader(CompileShader(vs_5_0, FinalColor_vs_preprocessor()));
		SetGeometryShader(NULL);
		SetPixelShader(CompileShader(ps_5_0, DCS_Forward_ps_preprocessor()));
		SetBlendState(FinalColor_BlendState, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xFFFFFFFF);

		#ifdef FinalColor_OnlayMode
		// For rendering onlay materials on top of currently rendered material.
		SetDepthStencilState(FinalColor_DepthStencilState_OnlayMode, 0xFFFFFFFF);
		#elif defined(FinalColor_NoShadowsMode)
		SetDepthStencilState(FinalColor_DepthStencilState_NoShadowsMode, FLATSHADOW_STENCIL_MASK);
		#else
		SetDepthStencilState(FinalColor_DepthStencilState, FinalColor_Stencil_Mask);
		#endif

		SetRasterizerState(FinalColor_RasterizerState);
	}
}

#endif
