//#define FINALCOLOR_DISABLE_DEPTH_TEST		- отключает depth test
//#define FINALCOLOR_DISABLE_ALPHA_BLEND    - отключает альфа блендинг
#include "inc/StencilUsage.hlsl"
#include "inc/CullMode.hlsl"
#include "inc/DepthFunc.hlsl"
#include "common/stencil.hlsl"
#include "common/dithering.hlsl"

#if defined(DCS_SHADING) && defined(EDGE)
	#undef FOG_ENABLE
#else
	#ifndef FOG_ENABLE
		#define FOG_ENABLE
	#endif
#endif

#ifndef CLOUD_SHADOWS
	#define CLOUD_SHADOWS
#endif

#ifndef LIGHT_TEXTURE
	#define LIGHT_TEXTURE
#endif

#ifndef METASHADER
	#define METASHADER
#endif

#ifdef USE_DCS_DEFERRED
	#include "deferred/GBuffer.hlsl"
	#define DCS_STENCIL_MASK	STENCIL_COMPOSITION_MASK|STENCIL_SURFACE_SHADOW_RECEIVER
	#include "inc/BuildGBufferFromColorOutput.hlsl"
#else
	#define DCS_STENCIL_MASK 0
#endif

#include "inc/FinalColor.hlsl"

// туман/атмосфера
#ifdef EDGE
	#ifdef DCS_SHADING
		#ifdef USE_DCS_DEFERRED
		#include "deferred/atmosphere.hlsl"
		#else
		#include "enlight/deprecated/atmInscatter.hlsl"
		#endif
	#else
		#include "inc/fog.hlsl"
	#endif
#else
	#include "common/context.hlsl"
	#ifdef USE_DCS_DEFERRED
		#include "deferred/atmosphere.hlsl"
	#else
		#include "enlight/deprecated/atmInscatter.hlsl"
	#endif
#endif

#ifdef CUSTOM_DEPTH
	#include "../lockonshaders/common/customDepth.hlsl"	
#endif

// #define SURFACELOD
// #define NEXTLOD_BLENDING
#include "terrain/inc/surfaceLodsBlendingConfig.hlsl"
#include "terrain/inc/surfaceLodsBlending.hlsl"

#if defined(OLD_LOD0_BLENDING) || defined(OLD_LOD1_BLENDING)
static const float transitionDistance = 0.9;
#endif

// mitka: хак для снега
#if defined(USE_DCS_DEFERRED) && defined(CAUCASUS_NOISE)
#include "terrain/inc/CaucasusNoise_ApplySnow.hlsl"
#endif

#include "../deferred/_HMD.hlsl"
#include "../deferred/_helper.hlsl"

// FinalColor Shading Model

void FinalColor_vs_preprocessor(
	out float4 output_positionWS : COLOR0,
#ifdef CLOUD_SHADOWS
	out float4 output_cloudShadowsUV : COLOR1,
#endif
	out float4 output_NDCposition : SV_POSITION
VS_INOUT_DECLARATION
	)
{
	TransformOutput transformOutput;
	
	TRANSFORM_NAME(
		transformOutput
		TM_INOUT_CALL
		);
	// dummy declaration
VS_DUMMY_DECLARATION;
	float4 P = VS_NAME(
		transformOutput
		VS_INOUT_CALL
		);

#ifdef FINALCOLOR_heightmapProject
	// project P to terrain if possible
	#ifndef EDGE_PREVIEW
	heightmapProject(P.xyz);
	#endif
#endif

	// kill z-fighting with sea

	//if( P.y<TerrainContext_seaLevel)
	//	P.y = TerrainContext_seaLevel;

#if !defined(EDGE_PREVIEW) && defined(OLD_LOD0_BLENDING) && defined(DISCARD_POLYGONS_BY_SEELEVEL) 
	if(P.y<TerrainContext_seaLevel+TerrainConfig_seaDiscardDepth)
		P /= 0;
#endif

#if !defined(SURFACE5) && defined(NEXTLOD_CLIPPING)
	blendSurfaceClipping(P);
#endif
		

//#ifdef WINDDEFORMER_ENABLE
//	ApplyWindDeformer(P, transformOutput.m[3].xyz);
//#endif

	float4 Prender = P;
#ifdef OLD_LOD1_BLENDING
	// LOD: move close verts down
	float dist = length(P - TerrainContext_cameraPos);

	dist = (dist-transitionDistance*TerrainContext_surfaceLevel0) / ( (1-transitionDistance)*TerrainContext_surfaceLevel0);
	Prender.y += -1 + lerp( -10, 0, saturate(dist));
#endif

#ifdef FINALCOLOR_VSBias
	float3 cd = TerrainContext_cameraPos - P;
	Prender.xyz += cd*FINALCOLOR_VSBias;
#endif

	// shading model
	output_NDCposition = mul(Prender, TerrainContext_VP);
	output_positionWS = P;

#ifdef CLOUD_SHADOWS
	output_cloudShadowsUV = mul(P+float4(TerrainContext_Origin, 0), SharedParams_cloudShadowsProj);
#endif

}

void FinalColor_ps_preprocessor(
	float4 positionWS : COLOR0,
#ifdef CLOUD_SHADOWS
	float4 cloudShadowsUV : COLOR1,
#endif
	float4 sv_pos : SV_Position,
#ifdef CUSTOM_DEPTH
	out float depth : SV_DEPTH,
#endif
#ifndef USE_DCS_DEFERRED
	out float4 color : SV_Target
#else
	out GBuffer gbuffer
#endif
#if defined(THIS_IS_TREE) && ENABLE_ATOC
  ,out uint coverageMask : SV_Coverage
#endif
PS_IN_DECLARATION

#ifdef USE_DCS_DEFERRED      
	#if USE_SV_SAMPLEINDEX
		, uint sv_sampleIndex : SV_SampleIndex  
	#endif
#endif
	)
{
#if defined(THIS_IS_TREE) && ENABLE_ATOC
	coverageMask = 255;
#endif  
	float4 NDCposition = mul(positionWS, TerrainContext_VP);

#if defined(SURFACE5) && defined(NEXTLOD_CLIPPING) && defined(USE_blendSurfaceClippingPS)
	float alphaDistance = 0;
	int visible = blendSurfaceClippingPS(positionWS, alphaDistance);
	//color = float4(visible*0.5, 0, 0, 1); return;
	if (visible == 0)
		discard;
#endif

	// дистанция
	float dist = length( TerrainContext_cameraPos-positionWS.xyz);
#ifdef OLD_LOD0_BLENDING
	float distanceBlend = TerrainContext_surfaceLevel0*(1-transitionDistance);
	float alphaDistance = lerp( 0, 1, saturate( (TerrainContext_surfaceLevel0-dist)/distanceBlend ));
	if(alphaDistance==0)
		discard;
#endif

#ifdef OLD_LOD1_BLENDING
	float distanceBlend = transitionDistance*TerrainContext_surfaceLevel0;
	float alphaDistance = step( distanceBlend, dist );
	if(alphaDistance==0)
		discard;
	//color = float4(alphaDistance, alphaDistance, alphaDistance, 1); return;
#endif

#ifdef CUSTOM_DEPTH
	float zNear = TerrainContext_P[3][2] / (TerrainContext_P[2][2]-1.0);
	float zFar = TerrainContext_P[3][2] / (1.0+TerrainContext_P[2][2]);
	depth=CustomDepth(NDCposition, zNear, zFar);
#endif

	ColorOutput output;
	ColorOutput_Init(output);
	// вызов пиксельного шейдера
	PS_NAME(
		output, 
		positionWS 
		PS_IN_CALL);

#if defined(THIS_IS_TREE) && ENABLE_ATOC
  if (COVERAGE_MASK > 1)
  {
    coverageMask = AToC(output.fAlpha);  
    output.fAlpha = 1;
  }
#endif
    
#ifdef USE_DITHERING
	if (dither_ordered8x8(sv_pos.xy) >= output.fAlpha)
		discard;
	output.fAlpha = 1;
#endif


// mitka: хак для снега в кавказе
#if defined(USE_DCS_DEFERRED) && defined(CAUCASUS_NOISE)
	CaucasusNoise_ApplySnow(output);
#endif

#ifndef UNUSE_COMPOSITE_FINAL_COLOR
	#ifdef PBR_TEXTURES
		output.fRoughness = clamp(output.fRoughness, MIN_ROUGHNESS, MAX_ROUGHNESS);
	#endif
#endif

#if defined(RENDER_VARYHOUSE_COMPONENTS) && !defined(THIS_IS_VARYHOUSE_MATERIAL)
	discard;
#endif
#if defined(RENDER_SURFACE_COMPONENTS) && !defined(THIS_IS_SURFACE_MATERIAL)
	discard;
#endif

#ifdef USE_DCS_DEFERRED

#if ( defined(SURFACE5) && defined(NEXTLOD_CLIPPING) && defined(USE_blendSurfaceClippingPS) ) || defined(OLD_LOD0_BLENDING) ||  defined(OLD_LOD1_BLENDING)
	output.fAlpha = min( output.fAlpha, alphaDistance);
	if(output.fAlpha==0)
		discard;
#endif
	gbuffer = BuildGBufferFromColorOutput(
		output, positionWS, NDCposition, sv_pos.xy
#if USE_SV_SAMPLEINDEX
		,sv_sampleIndex
#endif
		);

	return;
#else

#if (defined(RENDER_SURFACE_COMPONENTS) || defined(RENDER_VARYHOUSE_COMPONENTS) || defined(RENDER_RUNWAY_COMPONENTS)) && !defined(UNUSE_COMPOSITE_FINAL_COLOR)
	color = float4(output.vDiffuse, 1);
	return;
#endif

#if !defined(UNUSE_COMPOSITE_FINAL_COLOR)
// for tests
if (output.fMaterialID < 0)
{
	color = float4(output.vDiffuse, 1);
	return;
}
#else
	color = output.color;
	return;
#endif

#ifndef CLOUD_SHADOWS
	float4 cloudShadowsUV = float4(0, 0, 0, 1);
#endif

	// shading model
	color = FinalColor(
		output, 
		positionWS,
		NDCposition,  
		cloudShadowsUV,
		false, 
		false);

#if defined(RENDER_SHADING_COMPONENTS)
	if(color.a==0)
		discard;
	return;
#endif

#ifdef OLD_LOD0_BLENDING
	color.a = min( color.a, alphaDistance);
	if(color.a==0)
		discard;
#if !defined(EDGE_PREVIEW) && !defined(SURFACE5)
	// отрезать дно
	float maxDepth = lerp(TerrainContext_seaLevel-2, TerrainContext_seaLevel, saturate( (dist-5000)/5000) );
	if(positionWS.y <= maxDepth)
		discard;
#endif
#endif

#if defined(SURFACE5) && defined(NEXTLOD_CLIPPING) && defined(USE_blendSurfaceClippingPS)
	color.a = min( color.a, alphaDistance);
	if(color.a==0)
		discard;
#endif

	// туман
#if defined(DCS_SHADING) || !defined(EDGE) 
//	if( SharedParams_atmosphereAvailable)
		color.rgb = atmosphereApply(TerrainContext_cameraPos.xyz, positionWS.xyz/positionWS.w, NDCposition, color.rgb);
#else
	#ifdef FOG_ENABLE
	{
		float3 pos = positionWS.xyz / positionWS.w;
		fogApply(TerrainContext_cameraPos, pos, NDCposition, color.rgb); 
	}
	#endif
#endif

#endif

}

#ifndef FINALCOLOR_OVERRIDE_TECHNIQUE

#ifndef FINALCOLOR_DISABLE_ALPHA_BLEND

	#ifdef USE_DCS_DEFERRED
		BlendState FinalColor_BlendState
		{
			BlendEnable[0] = TRUE;		// blendig of normals
			SrcBlend[0] = SRC_ALPHA;
			DestBlend[0] = INV_SRC_ALPHA;
			BlendOp[0] = ADD;
		#ifdef DISCARD_NORMAL
			RenderTargetWriteMask[0] = 0x00; 
		#else
			RenderTargetWriteMask[0] = 0x0f;
		#endif

			BlendEnable[1] = TRUE;
			SrcBlend[1] = SRC_ALPHA;
			DestBlend[1] = INV_SRC_ALPHA;
			BlendOp[1] = ADD;
			RenderTargetWriteMask[1] = 0x0f;

			BlendEnable[2] = true;
			SrcBlend[2] = SRC_ALPHA;
			DestBlend[2] = INV_SRC_ALPHA;
			BlendOp[2] = ADD;
			RenderTargetWriteMask[2] = 0x0f;

			BlendEnable[3] = true;
			SrcBlend[3] = SRC_ALPHA;
			DestBlend[3] = INV_SRC_ALPHA;
			BlendOp[3] = ADD;
			RenderTargetWriteMask[3] = 0x0f;
		};
	#else
		BlendState FinalColor_BlendState
		{
			BlendEnable[0] = TRUE;
			BlendEnable[1] = TRUE;
			SrcBlend = SRC_ALPHA;
			DestBlend = INV_SRC_ALPHA;
			BlendOp = ADD;
			SrcBlendAlpha = SRC_ALPHA;
			DestBlendAlpha = INV_SRC_ALPHA;//ZERO;
			BlendOpAlpha = ADD;
			RenderTargetWriteMask[0] = 0x0f; //RED | GREEN | BLUE | ALPHA
		};
	#endif

#else
	BlendState FinalColor_BlendState
	{
		BlendEnable[0] = FALSE;
		BlendEnable[1] = FALSE;
		BlendEnable[2] = FALSE;
		BlendEnable[3] = FALSE;
	};
#endif

#ifndef FINALCOLOR_DepthEnable
	#ifndef FINALCOLOR_DISABLE_DEPTH_TEST
		#define FINALCOLOR_DepthEnable TRUE
	#else
		#define FINALCOLOR_DepthEnable FALSE
	#endif
#endif

#ifndef FINALCOLOR_DepthWriteMask
	#define FINALCOLOR_DepthWriteMask All
#endif

DepthStencilState FinalColor_DepthStencilState
{
	DepthEnable        = FINALCOLOR_DepthEnable;
	DepthWriteMask     = FINALCOLOR_DepthWriteMask;
	DepthFunc          = DEPTH_FUNC;

	StencilEnable      = TRUE;
#if defined(FINALCOLOR_VSBias) && defined(USE_DCS_DEFERRED)
	StencilReadMask    = DCS_STENCIL_MASK;
	StencilWriteMask   = SURFACE_STENCIL_MASK;

	FrontFaceStencilFunc = EQUAL;
	FrontFaceStencilFail = KEEP;
	FrontFaceStencilPass = REPLACE;
	BackFaceStencilFunc  = EQUAL;
	BackFaceStencilFail  = KEEP;
	BackFaceStencilPass  = REPLACE;
#else
	StencilReadMask    = SURFACE_STENCIL_MASK;
	StencilWriteMask   = SURFACE_STENCIL_MASK | DCS_STENCIL_MASK;

	FrontFaceStencilFunc = ALWAYS;
	FrontFaceStencilFail = KEEP;
	FrontFaceStencilPass = REPLACE;
	BackFaceStencilFunc  = ALWAYS;
	BackFaceStencilFail  = KEEP;
	BackFaceStencilPass  = REPLACE;
#endif
};

#define FinalColor_Stencil_Mask 0

#ifndef DCS_FinalColor_Stencil_Value
	// все что торчит из земли шейдим как модели
	#if !defined(EDGE) && defined(USE_DCS_DEFERRED) && defined(THIS_IS_BUILDING)
		#define DCS_FinalColor_Stencil_Value STENCIL_COMPOSITION_MODEL
	#else
		#define DCS_FinalColor_Stencil_Value 0
	#endif

#endif

// mitka: STENCIL_COMPOSITION_MODEL for all materials
//#ifdef PBR_TEXTURES
//	#undef DCS_FinalColor_Stencil_Value
//	#define DCS_FinalColor_Stencil_Value STENCIL_COMPOSITION_MODEL
//#endif

#ifndef FinalColor_OnlayMode_StencilFunc
	#define FinalColor_OnlayMode_StencilFunc ALWAYS
#endif

DepthStencilState FinalColor_DepthStencilState_OnlayMode
{
	DepthEnable        = FINALCOLOR_DepthEnable;
	DepthWriteMask     = FINALCOLOR_DepthWriteMask;
	DepthFunc          = DEPTH_FUNC;

	StencilEnable      = TRUE;
	StencilReadMask    = SURFACE_STENCIL_MASK;
	StencilWriteMask   = SURFACE_STENCIL_MASK | DCS_STENCIL_MASK;

	FrontFaceStencilFunc = FinalColor_OnlayMode_StencilFunc;
	FrontFaceStencilPass = REPLACE;
	FrontFaceStencilFail = KEEP;
	BackFaceStencilFunc  = FinalColor_OnlayMode_StencilFunc;
	BackFaceStencilFail  = KEEP;
	BackFaceStencilPass  = REPLACE;
};

DepthStencilState FinalColor_DepthStencilState_NoShadowsMode
{
	DepthEnable        = FINALCOLOR_DepthEnable;
	DepthWriteMask     = FINALCOLOR_DepthWriteMask;
	DepthFunc          = DEPTH_FUNC;

	StencilEnable      = TRUE;
	StencilReadMask    = FLATSHADOW_STENCIL_MASK | SURFACE_STENCIL_MASK;
    StencilWriteMask   = FLATSHADOW_STENCIL_MASK | SURFACE_STENCIL_MASK | DCS_STENCIL_MASK;

	FrontFaceStencilFunc = ALWAYS;
	FrontFaceStencilFail = REPLACE;
	FrontFaceStencilPass = REPLACE;
	BackFaceStencilFunc  = ALWAYS;
	BackFaceStencilPass  = REPLACE;
	BackFaceStencilFail  = REPLACE;
};

// FINALCOLOR_DepthBias, FINALCOLOR_DepthBiasClamp, FINALCOLOR_SlopeScaledDepthBias
#ifndef FINALCOLOR_DepthBias
	#define FINALCOLOR_DepthBias 0
#endif
#ifndef FINALCOLOR_DepthBiasClamp
	#define FINALCOLOR_DepthBiasClamp 0
#endif
#ifndef FINALCOLOR_SlopeScaledDepthBias
	#define FINALCOLOR_SlopeScaledDepthBias 0
#endif

RasterizerState FinalColor_RasterizerState
{
	CullMode = FINALCOLOR_CullMode;
	FillMode = Solid;
	MultisampleEnable = TRUE;

	DepthBias = FINALCOLOR_DepthBias;
	DepthBiasClamp = FINALCOLOR_DepthBiasClamp;
	SlopeScaledDepthBias = FINALCOLOR_SlopeScaledDepthBias;
};

#endif //FINALCOLOR_OVERRIDE_TECHNIQUE

technique10 FinalColor_Technique
{
	pass P0
	{
		SetVertexShader(CompileShader(vs_4_1, FinalColor_vs_preprocessor()));
		SetGeometryShader(NULL);
		SetPixelShader(CompileShader(ps_5_0, FinalColor_ps_preprocessor()));
		SetBlendState(FinalColor_BlendState, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xFFFFFFFF);

		#ifdef FinalColor_OnlayMode
		// For rendering onlay materials on top of currently rendered material.
		SetDepthStencilState(FinalColor_DepthStencilState_OnlayMode, FinalColor_Stencil_Mask | DCS_FinalColor_Stencil_Value | STENCIL_SURFACE_SHADOW_RECEIVER | SURFACE_STENCIL_MASK);
		#elif defined(FinalColor_NoShadowsMode)
		SetDepthStencilState(FinalColor_DepthStencilState_NoShadowsMode, FLATSHADOW_STENCIL_MASK | DCS_FinalColor_Stencil_Value | STENCIL_SURFACE_SHADOW_RECEIVER);
		#else
        SetDepthStencilState(FinalColor_DepthStencilState, FinalColor_Stencil_Mask | DCS_FinalColor_Stencil_Value | STENCIL_SURFACE_SHADOW_RECEIVER);
		#endif

		SetRasterizerState(FinalColor_RasterizerState);
	}
}


