#include "inc/StencilUsage.hlsl"
#include "inc/flatshadow.hlsl"
#include "terrain/inc/heightmapSampler.hlsl"
#include "inc/DepthFunc.hlsl"
#include "terrain/inc/RENDER_SHADING_COMPONENTS.hlsl"

// туман
#ifdef EDGE
	#include "inc/fog.hlsl"
#else
	#include "enlight/skyCommon.hlsl"
	#include "deferred/shadows.hlsl"
	#include "common/States11.hlsl"
	#ifdef FOG_ENABLE
	#include "common/fogCommon.hlsl"
	#endif
#endif

#ifndef DCS_SHADING
	#define FLATSHADOWS_CUSTOMDEPTH
#endif

#include "../deferred/_HMD.hlsl"
#include "../deferred/_helper.hlsl"

// FlatShadows Shading Model

//#ifdef RENDER_SHADING_COMPONENTS
//#endif

void FlatShadows_vs_preprocessor(
	out float4 output_positionWS : COLOR0,
	out float1 output_distanceToLand : COLOR1,
	out float4 output_NDCposition : SV_POSITION
VS_INOUT_DECLARATION
	)
{
	TransformOutput transformOutput;
	TRANSFORM_NAME(
		transformOutput
		TM_INOUT_CALL
		);
	VS_DUMMY_DECLARATION;
	float4 P = VS_NAME(
		transformOutput
		VS_INOUT_CALL
		);
	// shading model
	float4 P0 = mul(float4(0, 0, 0, 1), transformOutput.m);
	float distanceToLand = P.y-P0.y;
	P.xyz = vs_flatshadow( P.xyz, float3(0, 1, 0), distanceToLand);

	// project P to terrain if possible
#ifndef EDGE_PREVIEW
	heightmapProject(P.xyz);
	P.y = max(P.y, TerrainContext_seaLevel);
#endif
	float4 Pbias = P;

#ifdef FLATSHADOWS_CUSTOMDEPTH
	Pbias.y += 0.1;
#else
	float3 cd = TerrainContext_cameraPos - P;
	Pbias.xyz += cd*0.05;
	//Pbias.xyz += normalize(cd)*(2 + length(cd)*0.05);
#endif

	output_NDCposition = mul(Pbias, TerrainContext_VP);
	output_distanceToLand = distanceToLand;
	output_positionWS = P;
}

#ifdef CLIPMAP_SHADOW_MAP
#include "terrain/inc/clipmap.hlsl"
CLIPMAP(shadowmap)
#endif

void FlatShadows_ps_preprocessor(
	float4 positionWS : COLOR0,
	float1 distanceToLand : COLOR1,
	float4 NDCposition : SV_Position,
	out float4 color : SV_Target0
	#ifdef FLATSHADOWS_CUSTOMDEPTH
	,out float depth : SV_Depth
	#endif
#if ENABLE_ATOC
  ,out uint coverageMask : SV_Coverage
#endif
PS_IN_DECLARATION
	)
{
#if ENABLE_ATOC
	coverageMask = 255;
#endif
#ifdef RENDER_SHADING_COMPONENTS
	if (SharedParams_shadingHint != RSC_FLATSHADOWS)
		discard;
#endif
#ifdef RENDER_SURFACE_COMPONENTS
	discard;
#endif


	clip(distanceToLand);

	// вызов пиксельного шейдера
	float shadow = PS_NAME(
		positionWS 
		PS_IN_CALL);
#ifdef EDGE
	shadow *= SharedParams_shadowsPower;
#endif

#if ENABLE_ATOC
  if (COVERAGE_MASK > 1)
  {
    coverageMask = AToC(shadow);
  }
#endif  

	float4 positionNDC = mul(positionWS, TerrainContext_VP);

#ifdef USE_DCS_DEFERRED
	float dist = positionNDC.z / positionNDC.w;
	if(dist > FlatShadowDistance[2])
		discard;

	color = float4(0, 0, 0, 1);

#else

	#ifdef CLIPMAP_SHADOW_MAP
	float landshadow = SAMPLE_CLIPMAP_SHADOWMAP(shadowmap, positionWS.xyz, SharedParams_sunDirectionWS);
	landshadow = saturate((landshadow-0.5)*5);
	landshadow = 1 - landshadow;
	shadow *= landshadow;
	#endif

#ifdef EDGE
	#if !defined(USE_DCS_DEFERRED) && !defined(DCS_SHADING) 
		// если включены облака stratus(сплошная облачность), добавляем от них искусственную тень по высоте
		float heightIntensity = smoothstep(TerrainContext_overcastMinMaxIntensity.x, TerrainContext_overcastMinMaxIntensity.y, positionWS.y);
		float overcastShadowIntensity = lerp(0.5, 1.0, heightIntensity);
		shadow *= lerp(overcastShadowIntensity, 1.0, TerrainContext_overcastMinMaxIntensity.z);
	#endif
	
	color = float4(0, 0, 0, shadow);
#else
	// дистанция?
	float dist = positionNDC.z / positionNDC.w;
	float3 pos = positionWS.xyz / positionWS.w;

	float kDiffuse = TerrainContext_materialParams[0].kDiffuse;
	float kAmbient = TerrainContext_materialParams[0].kAmbient;
	float kShadows = TerrainContext_surface_landShadowsPower;

	float shadowPowered = lerp(1 - kShadows, 1, shadow);

	float3 ambient = kAmbient*AmbientLight(float3(0, 1, 0));
	float sunDot = pow(LightDot(float3(0, 1, 0), SharedParams_sunDirectionWS), 0.8); //освещенность солнцем [0;1]
	float sunDotShadowPowered = sunDot*shadowPowered; //освещенность солнцем c корректированой тенью [0;1]
	float3 diffuse = kDiffuse*sunDot;

	float shad = FlatShadowValue(dist);
	if (shad<0.001)
		discard;

	color.rgb = ambient / diffuse;
	color.rgb = lerp(float3(1, 1, 1), color.rgb, TerrainContext_surface_landShadowsPower);	
	color.rgb = lerp(float3(1, 1, 1), color.rgb, shadow);
	color.rgb = lerp(float3(1, 1, 1), color.rgb, sqrt(shad));

	float d = length(pos);
	float attL = fogCalcAttenuation(fogCoeff.z, fogCoeff.w, fogCameraHeight, d, pos.y / d);
	color.rgb = lerp(float3(1, 1, 1), color.rgb, attL*attL);

	color.a = 1;

	//color = FlatShadowColorWithFog(dist, pos);

#endif

	// z and depth bias
#ifdef FLATSHADOWS_CUSTOMDEPTH
	positionNDC = mul(positionWS, TerrainContext_V);
	positionNDC.z += 5;
	if(positionNDC.z>0) positionNDC.z = 0;
	positionNDC = mul(positionNDC, TerrainContext_P);
	float zbias = -0.01;
	depth = ( positionNDC.z+zbias)/positionNDC.w;
	depth = saturate(depth);
#endif

	// туман
#ifndef DCS_SHADING
	#ifdef FOG_ENABLE
	{
		float3 pos = positionWS.xyz / positionWS.w;
		fogApply(TerrainContext_cameraPos, pos, positionNDC, color.rgb); 
	}
	#endif
#else
#endif

#endif // USE_DCS_DEFERRED

}

BlendState FlatShadows_BlendState
{
	BlendEnable[0] = TRUE;
	BlendEnable[1] = TRUE;
	SrcBlend = SRC_ALPHA;
	DestBlend = INV_SRC_ALPHA;
	BlendOp = ADD;
	SrcBlendAlpha = SRC_ALPHA;
	DestBlendAlpha = INV_SRC_ALPHA;//ZERO;
	BlendOpAlpha = ADD;
	RenderTargetWriteMask[0] = 0x0f; //RED | GREEN | BLUE | ALPHA
};
DepthStencilState FlatShadows_DepthStencilState
{
	DepthEnable          = TRUE;
	DepthWriteMask       = ZERO;
	DepthFunc            = DEPTH_FUNC;

	StencilEnable        = TRUE;
	StencilReadMask      = FLATSHADOW_STENCIL_MASK;
	StencilWriteMask     = FLATSHADOW_STENCIL_MASK;

	FrontFaceStencilFunc = NOT_EQUAL;
	FrontFaceStencilPass = REPLACE;
	FrontFaceStencilFail = KEEP;
	
	BackFaceStencilFunc  = NOT_EQUAL;
	BackFaceStencilPass  = REPLACE;
	BackFaceStencilFail  = KEEP;
};
RasterizerState FlatShadows_RasterizerState
{
	CullMode = None;
	FillMode = Solid;
	MultisampleEnable = TRUE;
#if !defined(EDGE) || defined(USE_INVERSE_PROJ)
	DepthBias = 200;
#endif
//	SLOPESCALEDDEPTHBIAS = -10000;
};

technique10 FlatShadows_Technique
{
	pass P0
	{
		SetVertexShader(CompileShader(vs_4_1, FlatShadows_vs_preprocessor()));
		SetGeometryShader(NULL);
		SetPixelShader(CompileShader(ps_4_1, FlatShadows_ps_preprocessor()));
//		SetBlendState(FlatShadows_BlendState, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xFFFFFFFF);
#ifdef EDGE
		SetBlendState(FlatShadows_BlendState, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xFFFFFFFF);
#else
		SetBlendState(enableFlatShadowAlphaBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xFFFFFFFF);
#endif
		SetDepthStencilState(FlatShadows_DepthStencilState, 0xFFFFFFFF);
		SetRasterizerState(FlatShadows_RasterizerState);
	}
}
