#include "Speedtree5.1.hlsl"

#include "../common/context.hlsl"
#include "../deferred/_HMD.hlsl"
#include "../deferred/_helper.hlsl"

Texture2D material::normalTexture;
int material::FORMATOFnormalTexture;

#ifdef USE_3D_AO
Texture3D material::AOTexture;
#define MPS_MATRIX_AOMatrix		0
#endif

#ifdef USE_PALETTE
Texture2D material::paletteTexture;
#endif

// 0 = Tree, Bilboard: paint only close to base color
// 1 = Tree: paint only close to base color. Bilboard: paint fully
// 2 = Tree, Bilboard: paint fully
#define PAINT_ONLY_BASE_COLOR_MODE 0


//----------------------------------------------------------------------------------------------------------------------------------------------------

float4 @default_vs(
	TransformOutput transform,	

	float3 geometry::P,
	float3 geometry::N,
	int    geometry::options,
	float2 geometry::diffuseTexCoord,
	float4 geometry::lodTransition,
	float1 geometry::geometryType,
	float3 geometry::tangent,

	int  geometry::materialParamsIndex,

	// outputs
	out float fade2bilboard,
	out float gamma,

	out float3 normal,
	out float3 tangent,
	out float3 binormal, 
	out float2 colorTexCoord,
	out float1 blendFactor,
	out float1 seed,
	out float4 localpos
#ifdef USE_3D_AO
,	out float3 AOTexCoord
#endif
#ifdef USE_PALETTE
,	out float2 paletteUV
,	out float4 baseColor
#endif
,	out uint geometryType
	)
{
	float3 cameraToTree = -normalize(SharedParams_cameraPos.xyz - mul(float4(0, 0, 0, 1), transform.m).xyz);
	float4 position = commonVS(transform,
		cameraToTree, TerrainContext_cameraPos, TerrainContext_cameraDir,
		geometry::P, geometry::N, geometry::options, geometry::diffuseTexCoord, geometry::lodTransition, geometry::geometryType, geometry::tangent, geometry::materialParamsIndex,
		fade2bilboard, normal, tangent, binormal, colorTexCoord, blendFactor, geometryType, localpos);

	MaterialParamsStruct mps = materialparamsarray__materialParams[geometry__materialParamsIndex];
	gamma = mps.floats[MPS_FLOAT_GAMMA];
	
	int paletteVVals[6] = {
		mps.ints[MPS_INT_paletteVBranches],
		mps.ints[MPS_INT_paletteVFronds],
		mps.ints[MPS_INT_paletteVLeafs],
		mps.ints[MPS_INT_paletteVFacingLeafs],
		mps.ints[MPS_INT_paletteVBillboardVertical],
		mps.ints[MPS_INT_paletteVBillboardHorizontal]
	};

	float4 baseColorVals[6] = {
		mps.floats4[MPS_FLOAT4_baseColorBranches],
		mps.floats4[MPS_FLOAT4_baseColorFronds],
		mps.floats4[MPS_FLOAT4_baseColorLeafs],
		mps.floats4[MPS_FLOAT4_baseColorFacingLeafs],
		mps.floats4[MPS_FLOAT4_baseColorBillboardVertical],
		mps.floats4[MPS_FLOAT4_baseColorBillboardHorizontal]
	};

#ifdef USE_3D_AO
	AOTexCoord = saturate(mul(localpos, mps.matrixes[MPS_MATRIX_AOMatrix]));
#endif

	seed = transform.seed;

	// Palette
#ifdef USE_PALETTE
	float2 paletteTextureDims = float2(256, 4);
	material::paletteTexture.GetDimensions(paletteTextureDims.x, paletteTextureDims.y);
	paletteUV.x = frac(transform.seed);
	paletteUV.x *= paletteTextureDims.x;
	paletteUV.x = floor(paletteUV.x) + 0.5;
	paletteUV.y = floor(paletteVVals[geometryType]) + 0.5;

	baseColor = baseColorVals[geometryType];
#endif

	return position;
}


//----------------------------------------------------------------------------------------------------------------------------------------------------
void @decomposed_ps(
	inout ColorOutput output,
	inout float4 positionWS,

	float fade2bilboard,
	float gamma,

	float3 normal,
	float3 tangent,
	float3 binormal,
	float2 colorTexCoord,
	float1 blendFactor,
	float1 seed,
	float4 localpos
#ifdef USE_3D_AO
,	float3 AOTexCoord
#endif
#ifdef USE_PALETTE
,	float2 paletteUV
,	float4 baseColor
#endif
,	uint geometryType
	)
{
	// normalmap
	//normal = normalToCamera(positionWS.xyz, normal);
	float3 ps_normal = normal;
	float4 surfacecolor = float4(1, 0, 0, 1);

	surfacecolor = material::diffuseTexture.Sample(WrapSampler, colorTexCoord);
	float4 normalTexColor = material::normalTexture.Sample(WrapSampler, colorTexCoord);

	float3 normalmap = normalTexColor.xyz;
#ifdef NO_BC3N
	normalmap = NormalFromColor(normalTexColor, 0);
#else
	normalmap = NormalFromColor(normalTexColor, material::FORMATOFnormalTexture);
#endif

	float3x3 tangentSpace = float3x3(normalize(tangent), normalize(binormal), normalize(normal));
	ps_normal = mul(normalmap, tangentSpace);

	surfacecolor.rgb = pow(surfacecolor.rgb, 1 / gamma);

#ifdef USE_PALETTE
	float4 paletteColor = float4(0, 0, 0, 0);
	paletteColor = material::paletteTexture.Load(int3(paletteUV.x, paletteUV.y, 0));

	float baseColorLuminance = dot(baseColor.rgb, float3(1, 1, 1));
	float nearNoiseLuminance = dot(surfacecolor.rgb, float3(1, 1, 1));
	float intensivity = nearNoiseLuminance / baseColorLuminance;
	float4 baseColorCorrected = baseColor * intensivity;

	float colorDiff = dot(pow(abs(baseColorCorrected.rgb - surfacecolor.rgb), 0.5), float3(1, 1, 1));
	float maxDiff = 0.75;
	float diffFactor = diffFactor = saturate(1.0 - colorDiff / maxDiff); // If difference between surfacecolor and baseColor is too big - palette influence is zero
	#if PAINT_ONLY_BASE_COLOR_MODE == 1
		if(geometryType >= 4)
			diffFactor = 1; 
	#elif PAINT_ONLY_BASE_COLOR_MODE == 2
		diffFactor = 1;
	#endif

	surfacecolor.rgb = saturate(surfacecolor.rgb + (paletteColor.rgb - baseColor.rgb) * intensivity * diffFactor);
#endif

	output.vDiffuse = surfacecolor.rgb;

#ifdef USE_3D_AO
	float4 aoColor = material::AOTexture.Sample(WrapSampler, AOTexCoord);
	//aoColor = material::AOTexture.Sample(WrapSampler, float3(0.5, 0.5, 0.5));
	float aoVal = geometryType > 3.5 ? aoColor.g : aoColor.r;
	aoVal = geometryType > 4.5 ? aoColor.b : aoVal;
	output.fAO = 1 - aoVal;
#endif

	output.vNormal		= normalize(ps_normal);

#ifdef PBR_TEXTURES
	output.fRoughness = TREE_ROUGHNESS;
	output.fMetallic = TREE_METALLIC;
#else
	output.fSpecPower = 0.0f;
	output.fSpecFactor	= 0.0f;
#endif

	output.vSelfIllumination = float4(0, 0, 0, 0);
	output.fReflectance	= 0.0f;
	output.fAlpha = fade2bilboard > 0 ? fade2bilboard * blendFactor : 1;
	output.fMaterialID  = TREE_MATERIAL;
	output.fHeightOverSurface = localpos.y;

#if ENABLE_ATOC
  if (COVERAGE_MASK > 1)
  {
    surfacecolor.a -= fade2bilboard*(1 - blendFactor);
    output.fAlpha = (surfacecolor.a - TREE_ALPHA_TRESHOLD) / max(fwidth(surfacecolor.a), 0.0001) + 0.5;
    return;
  }
#endif
	// discard by alpha
	float threshold = TREE_ALPHA_TRESHOLD / pow(TREE_ALPHA_LOD_FACTOR, material::diffuseTexture.CalculateLevelOfDetail(WrapSampler, colorTexCoord));
	if (surfacecolor.a < threshold)
		discard;
}


//----------------------------------------------------------------------------------------------------------------------------------------------------
void @reflection_ps(
	inout ReflectionOutput output,
	inout float4 positionWS,
	float fade2bilboard,
	float2 colorTexCoord,
	float1 blendFactor
	)
{
	float4 surfacecolor = float4(1, 0, 0, 1);
	surfacecolor = material::diffuseTexture.Sample(WrapSampler, colorTexCoord);

	// TODO: add soft diffuse
	output.color.rgb = surfacecolor.rgb;
	output.color.a = 1;

	// discard by alpha
	if (surfacecolor.a - fade2bilboard*(1 - blendFactor) < TREE_ALPHA_TRESHOLD)
		discard;
}

//----------------------------------------------------------------------------------------------------------------------------------------------------
#include "../inc/map.hlsl"

void @map_ps(
	inout MapOutput output,
	inout float4 positionWS,
	float fade2bilboard,
	float2 colorTexCoord,
	float1 blendFactor

)
{
	float alpha = material::diffuseTexture.Sample(WrapSampler, colorTexCoord).a;

	// discard by alpha
	if (alpha - fade2bilboard*(1 - blendFactor) < TREE_ALPHA_TRESHOLD)
		discard;

	output.color = getMapColor(3);
}

//----------------------------------------------------------------------------------------------------------------------------------------------------
void @flir_ps(
	inout FlirOutput output,
	inout float4 positionWS,
	float fade2bilboard,
	float gamma,
	float3 normal,
	float3 tangent,
	float3 binormal,
	float2 colorTexCoord,
	float1 blendFactor,
	float1 seed,
	float4 localpos
#ifdef USE_3D_AO
	, float3 AOTexCoord
#endif
#ifdef USE_PALETTE
	, float2 paletteUV
	, float4 baseColor
#endif
	, uint geometryType
)
{
	ColorOutput color;
	ColorOutput_Init(color);
	decomposed_ps(color, positionWS, 
		fade2bilboard,
		gamma,
		normal,
		tangent,
		binormal,
		colorTexCoord,
		blendFactor,
		seed,
		localpos
#ifdef USE_3D_AO
		,AOTexCoord
#endif
#ifdef USE_PALETTE
		,paletteUV
		,baseColor
#endif
		,geometryType);

	output.fDataType = geometryType < 0.5 ? TERRAIN_FLIR_TREE_BRANCHES : TERRAIN_FLIR_TREE_LEAVES;
	output.vDiffuse = color.vDiffuse;
	output.vNormal = color.vNormal;
	output.fAlpha = color.fAlpha;
}