#define SPEEDTREE_DISABLE_NORMAL
#include "common/context.hlsl" //TODO: remopve?
#include "Speedtree5.1.hlsl"

#include "../deferred/_HMD.hlsl"
#include "../deferred/_helper.hlsl"

// When rendering shadows camera pos is nonsence. But according to this issue https://jira.eagle.ru/browse/GS-71
// currently view matrix for shadows is broken (and so is TerrainContext_cameraDir), so we take camera pos during shadows rendering 
// assuming it is something close to validness.
// ... Or take sundir which should be equal to camera dir in view matrix
#define CAMERA_DIR_MODE 1

float4 @default_vs(
	TransformOutput transform,	
	// vertex streams
	float3 geometry::P,
	int    geometry::options,
	float2 geometry::diffuseTexCoord,
	float4 geometry::lodTransition,
	float1 geometry::geometryType,

	int  geometry::materialParamsIndex,

	// outputs
	out float fade2bilboard,

	out float2 colorTexCoord,
	out float1 blendFactor
	)
{
	MaterialParamsStruct mps = materialparamsarray::materialParams[geometry::materialParamsIndex];
	float4 localpos = 0;
	uint geometryType = 0;

#ifdef CAMERA_DIR_MODE
	#if CAMERA_DIR_MODE == 0
		float3 cameraToTree = -normalize(SharedParams_cameraPos.xyz - mul(float4(0, 0, 0, 1), transform.m).xyz);
		float3 cameraDir = -normalize(SharedParams_cameraPos.xyz - mul(float4(geometry::P, 1), transform.m).xyz);
	#elif CAMERA_DIR_MODE == 1
		float3 cameraToTree = SharedParams_sunDirectionWS;
		float3 cameraDir = SharedParams_sunDirectionWS;
	#endif
#else
	float3 cameraToTree = TerrainContext_cameraDir;
	float3 cameraDir = TerrainContext_cameraDir;
#endif

	float3 fakeCameraPos = mul(float4(geometry::P, 1), transform.m).xyz - cameraDir * 1000.0;
	float4 position = commonVS(transform, 
		cameraToTree, fakeCameraPos, cameraDir,
		geometry::P, geometry::options, geometry::diffuseTexCoord, geometry::lodTransition, geometry::geometryType, geometry::materialParamsIndex,
		fade2bilboard, colorTexCoord, blendFactor, geometryType, localpos);

	return position;
}

float @flatShadows_ps(
	inout float4 positionWS,
	float fade2bilboard,
	float2 colorTexCoord,
	float1 blendFactor
	)
{
	float4 surfacecolor = float4(1, 0, 0, 1);

#if ENABLE_ATOC
  float alpha = 1;
  if (COVERAGE_MASK > 1)
  {
    surfacecolor = material::diffuseTexture.SampleBias(WrapSampler, colorTexCoord, 1.5);
    surfacecolor.a -= fade2bilboard*(1 - blendFactor);
    alpha = (surfacecolor.a - TREE_ALPHA_TRESHOLD) / max(fwidth(surfacecolor.a), 0.0001) + 0.5;
  }
  else
  {
    surfacecolor = material::diffuseTexture.Sample(WrapSampler, colorTexCoord);
    // discard by alpha
    if (surfacecolor.a - fade2bilboard*(1 - blendFactor) < TREE_ALPHA_TRESHOLD)
      discard;
  }
  return alpha;
#else
	surfacecolor = material::diffuseTexture.Sample(WrapSampler, colorTexCoord);

	// discard by alpha
	if (surfacecolor.a - fade2bilboard*(1 - blendFactor) < TREE_ALPHA_TRESHOLD)
		discard;

	return 1.0;
#endif
}

void @shadowmap_ps(
	inout float4 positionWS,
	out float alpha,
	float fade2bilboard,
	float2 colorTexCoord,
	float1 blendFactor
	)
{
	float4 surfacecolor = float4(1, 0, 0, 1);
	surfacecolor = material::diffuseTexture.Sample(WrapSampler, colorTexCoord);

	alpha = fade2bilboard > 0 ? fade2bilboard * blendFactor : 1;

	// discard by alpha
	float threshold = TREE_ALPHA_TRESHOLD / pow(TREE_ALPHA_LOD_FACTOR, material::diffuseTexture.CalculateLevelOfDetail(WrapSampler, colorTexCoord));
	if (surfacecolor.a < threshold)
		discard;
}