#ifndef MODEL_FORWARD_DEF_PS_HLSL
#define MODEL_FORWARD_DEF_PS_HLSL

#include "../../deferred/_HMD.hlsl"

#include "functions/shading.hlsl"

#ifdef ENABLE_DEBUG_UNIFORMS
#include "common/color_table.hlsl"
#include "common/debug_uniforms.hlsl"
#endif

float4 emissivePS(VS_OUTPUT input, uniform int Flags)
{
	MaterialParams mp = calcMaterialParams(input, MP_DIFFUSE | MP_NORMAL);

	float4 finalColor = float4(GammaToLinearSpace(mp.diffuse.rgb) * selfIlluminationValue, mp.diffuse.a*mp.diffuse.a);
#ifdef BANO_MATERIAL
	calcBANOAttenuation(mp.diffuse.a*mp.diffuse.a, mp.normal, mp.toCamera, mp.camDistance, finalColor.a);
#endif
	finalColor.rgb *= 5*finalColor.a;

	if(!(Flags & F_DISABLE_ATMOSPHERE))	{
		finalColor.rgb *= GetAtmosphereTransmittance(gCameraPos.xyz, mp.pos);
	}

	return finalColor;
}

float4 forwardDefaultPS(VS_OUTPUT input, uniform int Flags, uniform int shadingModel)
{
	if(shadingModel == SHADING_EMISSIVE)
		return emissivePS(input, Flags);

	MaterialParams mp = calcMaterialParams(input);
#if ENABLE_SIMPLE_TRANS && BLEND_MODE == BM_TRANSPARENT && defined(DEF_MATERIAL) && !defined(METROUGH_MAP)
  {
    AtmosphereSample atm;
#if BLEND_MODE == BM_NONE //opaque geometry
    atm.sunColor = SampleSunRadiance(mp.pos.xyz, gSunDir);
#else 
    atm = SamplePrecomputedAtmosphere(0);
    atm.sunColor /= gSunIntensity;
#endif
    float4 finalColor = mp.diffuse * float4(atm.sunColor, 1);
    if(!(Flags & F_DISABLE_ATMOSPHERE))
      finalColor.rgb = applyAtmosphereLinear(gCameraPos.xyz, mp.pos, input.projPos, finalColor.rgb);  
    return finalColor;
  }
#endif
  
	float shadow = 1.0;
	float2 cloudShadowAO = 1.0;
	if(!(Flags & F_ABOVE_CLOUDS))
	{
		cloudShadowAO = SampleShadowClouds(mp.pos);
		shadow = cloudShadowAO.x;
	}
	if(!(Flags & F_DISABLE_SHADOWMAP))
		shadow = min(shadow, applyShadow(float4(mp.pos, input.projPos.z/input.projPos.w), mp.normal));

	float4 aorm = getAORMS(input, mp.specular, mp.decalMask);
	mp.diffuse.rgb = modifyAlbedo(mp.diffuse.rgb, albedoLevel, albedoContrast, aorm.x);

	AtmosphereSample atm;
#if BLEND_MODE == BM_NONE //opaque geometry
	atm.sunColor =  SampleSunRadiance(mp.pos.xyz, gSunDir);
#else 
	atm = SamplePrecomputedAtmosphere(0);
	atm.sunColor /= gSunIntensity;
#endif

	float4 finalColor;

#ifdef ENABLE_DEBUG_UNIFORMS
	if(PaintNodes == 1){
		mp.emissive = CoarseGammaToLinearSpace(color_table[NodeId % N_COLORS]);
	}
#endif

	if(shadingModel == SHADING_STANDARD){
#if USE_SEPARATE_AO
		float AO = aorm.x;
#else
		float AO = 1;
#endif

		if(Flags & F_DISABLE_ATMOSPHERE){
			finalColor = float4(ShadeCockpit(input.Position.xy, (Flags & F_COCKPIT_GI), atm.sunColor, mp.diffuse.rgb, mp.normal, aorm.y, aorm.z, mp.emissive, shadow, AO, cloudShadowAO, mp.toCamera, mp.pos, float2(1,aorm.w), true, mp.diffuse.a), mp.diffuse.a);
		}else{
		#if BLEND_MODE == BM_TRANSPARENT
			finalColor = float4(ShadeHDR(input.Position.xy, atm.sunColor, mp.diffuse.rgb, mp.normal, aorm.y, aorm.z, mp.emissive, shadow, AO, cloudShadowAO, mp.toCamera, mp.pos, float2(1, aorm.w), LERP_ENV_MAP, false, float2(0, 0), LL_TRANSPARENT), mp.diffuse.a);
		#else
			finalColor = float4(ShadeHDR(input.Position.xy, atm.sunColor, mp.diffuse.rgb, mp.normal, aorm.y, aorm.z, mp.emissive, shadow, AO, cloudShadowAO, mp.toCamera, mp.pos, float2(1, aorm.w), LERP_ENV_MAP, false, float2(0, 0), BLEND_MODE == BM_ADDITIVE ? LL_TRANSPARENT : LL_SOLID), mp.diffuse.a);
		#endif
		}
	}else{
		finalColor = float4(1, 0, 0, 1);
	}

	if(!(Flags & F_DISABLE_ATMOSPHERE))
	{
	#if BLEND_MODE == BM_NONE
		finalColor.rgb = applyAtmosphereLinear(gCameraPos.xyz, mp.pos, input.projPos, finalColor.rgb);
	#else
		finalColor.rgb = finalColor.rgb * atm.transmittance + atm.inscatter;
	#endif
	}

	return finalColor;
}

#endif